library neuro;

Neuro neuro = Neuro();

void signal(value) {
  neuro.emit(value);
}

class Neuro {
  List<Handler> _handlers = [];

  void emit<T>([var value]) {
    print("neuro.emit(value: $value)");
    _handlers.forEach((handler) {
      if (!handler.enabled) return;

      if (value != null && !handler.canHandle(value)) return;

      if (value == null && !handler.canHandleType<T>()) return;

      Future futureResult = handler.handle(value).catchError((error) {
        emit(error);
        return null;
      });

      if (futureResult == null) {
        return;
      }

      futureResult.then((output) {
        if (output != null) {
          if (handler.canHandle(output)) {
            handler.enabled = false;
            print("handler disabled because it returns own handle type");
          } else {
            emit(output);
          }
        }
      });
    });
  }

  void handle<T>(HandlerFunction<T> _handlerFunction) {
    print("neuro.handle()");
    _handlers.add(Handler<T>(_handlerFunction));
  }

  void clear() {
    print("neuro.reset()");
    _handlers.clear();
  }
}

class Handler<T> {
  HandlerFunction<T> _handlerFunction;

  Handler(this._handlerFunction);

  bool enabled = true;

  Future handle(var value) async {
    return await _handlerFunction(value);
  }

  bool canHandle(var value) {
    return value is T;
  }

  bool canHandleType<J>() {
    return J == T;
  }
}

typedef Future<dynamic> HandlerFunction<E>(E element);

